import json
import random
from typing import List, Optional, Dict, Union
from fix_busted_json import repair_json, largest_json
from pydantic_prompter.common import Message, logger
from pydantic_prompter.llm_providers.bedrock_base import BedRock
from pydantic_prompter.annotation_parser import AnnotationParser


class BedRockAnthropic(BedRock):
    def __init__(
        self,
        model_name: str,
        parser: AnnotationParser,
        model_settings: Optional[Dict] = None,
    ):
        super().__init__(model_name, parser)
        self.model_settings = model_settings or {
            "temperature": random.uniform(0, 1),
            "max_tokens": 8000,
            "stop_sequences": ["Human:"],
            "anthropic_version": "bedrock-2023-05-31",
        }

    def _build_prompt(self, messages: List[Message], params: Union[dict, str]):
        return "\n".join([m.content for m in messages])

    @property
    def _template_path(self):
        path = self.settings.template_paths.anthropic.replace(
            "{prompt_paths}", self.parser.prompts_path
        )
        return path

    @property
    def _stop_sequence(self):
        return "Human:"

    def format_messages(self, msgs: List[Message]) -> str:
        return "\n".join([m.content for m in msgs])

    @staticmethod
    def fix_messages(msgs: List[dict]) -> List[dict]:
        # merge messages if roles do not alternate between "user" and "assistant"
        fixed_messages = []
        for m in msgs:
            if m["role"] == "system":
                m["role"] = "user"
            if fixed_messages and fixed_messages[-1]["role"] == m["role"]:
                fixed_messages[-1]["content"] += f'\n\n{m["content"]}'
            else:
                fixed_messages.append(m)
        return fixed_messages

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:

        if scheme:
            system_message = f"""Act like a REST API that performs the requested operation the user asked according to guidelines provided.
                    Your response should be a valid JSON format, strictly adhering to the Pydantic schema provided in the pydantic_schema section. 
                    Stick to the facts and details in the provided data, and follow the guidelines closely.
                    Respond in a structured JSON format according to the provided schema.
                    DO NOT add any other text other than the requested JSON response. 

                    ## pydantic_schema:

                    {json.dumps(scheme, indent=4)}
                    
                    """
        else:  # return_type:
            system_message = f"""Act like a REST API that performs the requested operation the user asked according to guidelines provided.
                    Your response should be according to the format requested in the return_type section. 
                    Stick to the facts and details in the provided data, and follow the guidelines closely.
                    Respond in a structured JSON format according to the provided schema.
                    DO NOT add any other text other than the requested return_type response. 

                    ## return_type:

                    {return_type}
                    
"""

        final_messages = [m.model_dump() for m in messages]
        final_messages = self.fix_messages(final_messages)

        # Ensure stop_sequences and anthropic_version are always included
        body = {
            "system": system_message,
            "messages": final_messages,
            "stop_sequences": self.model_settings.get(
                "stop_sequences", [self._stop_sequence]
            ),
            "anthropic_version": self.model_settings.get(
                "anthropic_version", "bedrock-2023-05-31"
            ),
            **self.model_settings,
        }

        response = self._boto_invoke(json.dumps(body))
        response_text = response.get("body").read().decode()
        response_json = repair_json(largest_json(response_text))
        response_body = json.loads(response_json)

        logger.info(response_body)
        return response_body.get("content")[0]["text"]



import abc
import json
import random
from typing import List, Union
from jinja2 import Template
from fix_busted_json import repair_json
from pydantic_prompter.common import Message, logger
from pydantic_prompter.exceptions import BedRockAuthenticationError
from pydantic_prompter.llm_providers.base import LLM


class BedRock(LLM, abc.ABC):
    @staticmethod
    def clean_result(body: str):
        body = body.split("<json_schema>")[0]
        body = body.replace("</json>", "")
        body = body.replace("</str>", "")
        body = body.replace("</int>", "")
        body = body.replace("</bool>", "")
        body = body.replace("<json>", "")
        body = body.replace("<str>", "")
        body = body.replace("<int>", "")
        body = body.replace("<bool>", "")
        body = body.replace("```", "")
        if "{" in body and "}" in body:
            left = body.find("{")
            right = body.rfind("}")
            body = body[left : right + 1]  # noqa
        return body

    @property
    @abc.abstractmethod
    def _template_path(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _stop_sequence(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def format_messages(self, msgs: List[Message]) -> str:
        raise NotImplementedError

    def _build_prompt(self, messages: List[Message], params: Union[dict, str]):
        if "prompt_templates" not in self._template_path:
            logger.info(f"Using custom prompt from {self._template_path}")
        ant_template = open(self._template_path).read()
        if isinstance(params, dict):
            scheme_ = json.dumps(params, indent=4)
        else:
            scheme_ = params
        ant_msgs = self.format_messages(messages)
        template = Template(ant_template, keep_trailing_newline=True)
        content = template.render(schema=scheme_, question=ant_msgs).strip()
        return content

    def debug_prompt(self, messages: List[Message], scheme: Union[dict, str]) -> str:
        return self._build_prompt(messages, scheme)

    def _boto_invoke(self, body):
        try:
            logger.debug(f"Request body: \n{body}")
            import boto3
            from botocore.config import Config

            session = boto3.Session(
                aws_access_key_id=self.settings.aws_access_key_id,
                aws_secret_access_key=self.settings.aws_secret_access_key,
                aws_session_token=self.settings.aws_session_token,
                profile_name=self.settings.aws_profile,
                region_name=self.settings.aws_default_region,
            )
            config = Config(
                read_timeout=120,  # 2 minutes before timeout
                retries={
                    "max_attempts": 5,
                    "mode": "adaptive",
                },  # retry 5 times adaptivly
                max_pool_connections=50,  # allow for significant concurrency
            )
            client = session.client("bedrock-runtime", config=config)
            # execute the model
            response = client.invoke_model(
                body=body,
                modelId=self.model_name,
                accept="application/json",
                contentType="application/json",
            )
        except Exception as e:
            raise BedRockAuthenticationError(e)

        return response

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        content = self._build_prompt(messages, scheme or return_type)

        body = json.dumps(
            {
                "max_tokens_to_sample": 8000,
                "prompt": content,
                "stop_sequences": [self._stop_sequence],
                "temperature": random.uniform(0, 1),
            }
        )

        response = self._boto_invoke(body)
        response_text = response.get("body").read().decode()
        response_json = repair_json(response_text)
        response_body = json.loads(response_json)
        logger.info(response_body)
        return response_body.get("completion")



from typing import List, Union

from pydantic_prompter.annotation_parser import AnnotationParser
from pydantic_prompter.common import Message


class LLM:
    @staticmethod
    def clean_result(body: str):
        return body

    def __init__(self, model_name: str, parser: AnnotationParser):
        from pydantic_prompter.settings import Settings

        self.parser: AnnotationParser = parser
        self.settings = Settings()
        self.model_name = model_name

    def debug_prompt(self, messages: List[Message], scheme: Union[dict, str]):
        raise NotImplementedError

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        raise NotImplementedError



import json
import random
from typing import List, Union

from pydantic_prompter.common import Message, logger
from pydantic_prompter.exceptions import OpenAiAuthenticationError
from pydantic_prompter.llm_providers.base import LLM


class OpenAI(LLM):
    @staticmethod
    def to_openai_format(msgs: List[Message]):
        openai_msgs = [item.model_dump() for item in msgs]
        return openai_msgs

    def debug_prompt(self, messages: List[Message], scheme: dict) -> str:
        return json.dumps(self.to_openai_format(messages), indent=4, sort_keys=True)

    @staticmethod
    def _create_schema(scheme: str) -> dict:
        if scheme == "str":
            ret = "string"
        elif scheme == "int":
            ret = "integer"
        elif scheme == "bool":
            ret = "boolean"
        elif scheme == "float":
            ret = "float"
        else:
            raise

        simple = {
            "name": "Simple",
            "parameters": {
                "properties": {"res": {"title": "Res", "type": ret}},
                "required": ["res"],
                "title": "Simple",
                "type": "object",
            },
        }
        return simple

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        from openai import OpenAI, OpenAIError
        from openai import AuthenticationError, APIConnectionError

        if return_type:
            scheme = self._create_schema(return_type)

        _function_call = {
            "name": scheme["name"],
        }
        logger.debug(f"Openai Functions: \n [{scheme}]")
        logger.debug(f"Openai function_call: \n {_function_call}")
        messages_oai = self.to_openai_format(messages)
        try:
            client = OpenAI(api_key=self.settings.openai_api_key)
            chat_completion = client.chat.completions.create(
                model=self.model_name,
                messages=messages_oai,
                functions=[scheme],
                function_call=_function_call,
                temperature=random.uniform(0.3, 1.3),
            )
        except (AuthenticationError, APIConnectionError, OpenAIError) as e:
            raise OpenAiAuthenticationError(e)
        return chat_completion.choices[0].message.function_call.arguments



from pathlib import Path
from typing import Optional

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from dotenv import find_dotenv

_root_path = Path(__file__).resolve().parent


class TemplatePaths(BaseModel):
    anthropic: str = Field(
        default=_root_path.joinpath(
            "prompt_templates", "{prompt_paths}", "anthropic.jinja"
        ).as_posix()
    )
    cohere: str = Field(
        default=_root_path.joinpath(
            "prompt_templates", "{prompt_paths}", "cohere.jinja"
        ).as_posix()
    )
    llama2: str = Field(
        default=_root_path.joinpath(
            "prompt_templates", "{prompt_paths}", "llama2.jinja"
        ).as_posix()
    )


class Settings(BaseSettings):
    openai_api_key: Optional[str] = None
    template_paths: TemplatePaths = TemplatePaths()
    aws_default_region: str = "us-east-1"
    aws_profile: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None
    cohere_key: Optional[str] = None
    model_config = SettingsConfigDict(
        env_file=find_dotenv(), env_nested_delimiter="__", extra="ignore"
    )



import random
from typing import List, Union

from pydantic_prompter.common import Message, logger
from pydantic_prompter.exceptions import CohereAuthenticationError
from pydantic_prompter.llm_providers.bedrock_cohere import BedRockCohere


class Cohere(BedRockCohere):
    def clean_result(self, body: str):
        body = super().clean_result(body)
        body = body.replace("<json>", "")
        body = body.replace("<str>", "")
        body = body.replace("<int>", "")
        body = body.replace("<bool>", "")
        return body

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        try:
            import cohere

            co = cohere.Client(api_key=self.settings.cohere_key)
            content = self._build_prompt(messages, scheme or return_type)

            response = co.chat(
                message=content,
                temperature=random.uniform(0, 1),
            )
            logger.debug(f"Request body: \n{content}")

        except Exception as e:
            logger.warning(e)
            raise CohereAuthenticationError(e)

        answer = response.text
        logger.debug(f"Got answer: \n{answer}")

        return answer



import logging
from typing import Optional, List, Any, Dict

from pydantic import BaseModel

from pydantic_prompter.settings import Settings

logger = logging.getLogger("pydantic_prompter")
logger.addHandler(logging.NullHandler())

settings = Settings()


class Message(BaseModel):
    role: str
    content: str

    def __str__(self):
        return f"{self.role}: {self.content}"


class LLMDataAndResult(BaseModel):
    inputs: Dict[str, Any]
    messages: Optional[List[Message]] = None
    raw_result: Optional[str] = None
    clean_result: Optional[str] = None
    result: Optional[BaseModel] = None
    error: Optional[Any] = None



import json
import random
from typing import List, Union

from pydantic_prompter.common import Message, logger
from pydantic_prompter.llm_providers.bedrock_base import BedRock


class BedRockLlama2(BedRock):

    @property
    def _template_path(self) -> str:
        path = self.settings.template_paths.llama2.replace(
            "{prompt_paths}", self.parser.prompts_path
        )
        return path

    @property
    def _stop_sequence(self) -> str:
        return "</s>"

    def format_messages(self, msgs: List[Message]) -> str:
        output = []
        for msg in msgs:
            if msg.role == "system":
                output.append(f"<<SYS>> {msg.content} <</SYS>>")
            if msg.role == "assistant":
                output.append(f"{msg.content}")
            if msg.role == "user":
                output.append(f"[INST] {msg.content} [/INST]")
        return "\n".join(output)

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        content = self._build_prompt(messages, scheme or return_type)

        body = json.dumps(
            {
                "max_gen_len": 2048,
                "prompt": content,
                "temperature": random.uniform(0, 1),
            }
        )
        response = self._boto_invoke(body)
        response_body = json.loads(response.get("body").read().decode())
        logger.debug(response_body)
        return response_body.get("generation")



from typing import Type, Dict, Union
from pydantic_prompter.annotation_parser import AnnotationParser
from pydantic_prompter.common import logger
from pydantic_prompter.llm_providers.bedrock_anthropic import BedRockAnthropic
from pydantic_prompter.llm_providers.bedrock_cohere import BedRockCohere
from pydantic_prompter.llm_providers.bedrock_llama2 import BedRockLlama2
from pydantic_prompter.llm_providers.cohere import Cohere
from pydantic_prompter.llm_providers.openai import OpenAI
from pydantic_prompter.llm_providers.base import LLM

# Mapping of llm type and model_name prefixes to their respective classes
LLM_MODEL_MAP: Dict[str, Dict[str, Type[LLM]]] = {
    "openai": {
        "default": OpenAI,
    },
    "bedrock": {
        "anthropic": BedRockAnthropic,
        "cohere": BedRockCohere,
        "meta": BedRockLlama2,
    },
    "cohere": {
        "command": Cohere,
    },
}


def get_llm(
    llm: str,
    model_name: str,
    parser: AnnotationParser,
    model_settings: Union[dict, None] = None,
) -> LLM:
    if llm not in LLM_MODEL_MAP:
        raise ValueError(f"LLM type '{llm}' is not implemented")

    # Extract the prefix from the model name. Adjust this logic as necessary.
    model_prefix = model_name.split(".")[
        0
    ]  # Extract 'anthropic' from 'anthropic.claude-3-sonnet-20240229-v1:0'

    model_class = LLM_MODEL_MAP.get(llm, {}).get(model_prefix, None)

    if model_class is None:
        raise ValueError(
            f"Model prefix '{model_prefix}' for LLM type '{llm}' is not implemented"
        )

    logger.debug(f"Using {model_class.__name__} provider with model {model_name}")

    return model_class(model_name, parser, model_settings)



import abc
import json
from json import JSONDecodeError
from typing import Dict, Any

from pydantic import ValidationError, ConfigDict

from pydantic_prompter.common import logger, LLMDataAndResult
from pydantic_prompter.exceptions import (
    FailedToCastLLMResult,
)


class AnnotationParser:
    @classmethod
    def get_parser(cls, function) -> "AnnotationParser":
        # noinspection PyProtectedMember
        from pydantic._internal._model_construction import ModelMetaclass

        return_obj = function.__annotations__.get("return", None)

        if isinstance(return_obj, ModelMetaclass):
            logger.debug("Using PydanticParser")
            return PydanticParser(function)

        elif return_obj in [str, int, float, bool]:
            return SimpleStringParser(function)

        else:
            raise Exception("Please make sure you annotate return type using Pydantic")

    def __init__(self, function):
        self.return_cls = function.__annotations__["return"]

    @abc.abstractmethod
    def llm_return_type(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def llm_schema(self) -> dict:
        raise NotImplementedError

    @abc.abstractmethod
    def cast_result(self, llm_data: LLMDataAndResult):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def prompts_path(self):
        raise NotImplementedError


class PydanticParser(AnnotationParser):
    def llm_return_type(self) -> str:
        pass

    def __init__(self, function):
        super().__init__(function)
        self.return_cls.model_config = ConfigDict(coerce_numbers_to_str=True)

    @property
    def prompts_path(self):
        return "pydantic"

    @staticmethod
    def pydantic_schema(schema_def: Dict[str, Any]) -> Any:
        return {
            "name": schema_def["title"],
            "description": schema_def.get("description", ""),
            "parameters": schema_def,
        }

    def llm_schema(self) -> dict:
        return_scheme = self.return_cls.model_json_schema(mode="serialization")
        return self.pydantic_schema(return_scheme)

    def cast_result(self, llm_data: LLMDataAndResult) -> LLMDataAndResult:
        try:
            j = json.loads(llm_data.clean_result, strict=False)
            res = self.return_cls(**j)
            llm_data.result = res
        except (ValidationError, JSONDecodeError) as e:
            llm_data.error = FailedToCastLLMResult(e)
        return llm_data


class SimpleStringParser(AnnotationParser):
    def llm_schema(self) -> dict:
        pass

    @property
    def prompts_path(self):
        return "simple"

    def llm_return_type(self) -> str:
        return self.return_cls.__name__

    def cast_result(self, llm_data: LLMDataAndResult) -> LLMDataAndResult:
        res = llm_data.clean_result
        if "{" in res and "}" in res and '"res":' in res:
            try:
                res = json.loads(res, strict=False)["res"]
            except JSONDecodeError as e:
                llm_data.error = FailedToCastLLMResult(e)
        try:
            llm_data.result = self.return_cls(res)
        except ValueError as e:
            llm_data.error = FailedToCastLLMResult(e)
        return llm_data



from typing import List, Optional, Dict

from jinja2 import Template
from retry import retry

from pydantic_prompter.annotation_parser import AnnotationParser
from pydantic_prompter.common import logger, Message, LLMDataAndResult
from pydantic_prompter.exceptions import (
    ArgumentError,
    BadRoleError,
    Retryable,
)
from pydantic_prompter.llm_providers import get_llm
from pydantic_prompter.llm_providers.base import LLM


class _Pr:
    def __init__(self, function, llm: str, model_name: str, jinja: bool, model_settings: Optional[Dict] = None):
        self.jinja = jinja
        self.function = function
        self.parser = AnnotationParser.get_parser(function)
        self.llm = get_llm(llm=llm, parser=self.parser, model_name=model_name, model_settings=model_settings)

    @retry(tries=3, delay=1, logger=logger, exceptions=(Retryable,))
    def __call__(self, *args, **inputs):
        if args:
            raise ArgumentError("please use only kwargs")

        llm_data = LLMDataAndResult(inputs=inputs)
        llm_data.messages = self._parse_function_to_messages(**inputs)
        logger.debug(f"Calling with prompt:\n{self.build_string(**inputs)}")
        res: LLMDataAndResult = self.call_llm(llm_data)

        if res.error:
            logger.error(f"\n\n ----> START OF ERROR <---- ")
            logger.exception(llm_data.error)
            logger.error(f"\n\nError ----> \n\n{type(res.error)}: {res.error}")
            logger.error(f"\n\nLLM output ----> \n\n{res.raw_result}")
            logger.error(f"\n\nLLM clean output ----> \n\n{res.clean_result}")
            logger.error(f"\n\nPrompt ----> \n\n{self.build_string(**inputs)}")
            logger.error(f"\n\n ----> END OF ERROR <---- ")
            raise res.error
        return res.result

    def build_string(self, **inputs) -> str:
        msgs = self._parse_function_to_messages(**inputs)
        res = self.llm.debug_prompt(
            msgs, self.parser.llm_schema() or self.parser.llm_return_type()
        )

        return res

    def _parse_function_to_messages(self, **inputs) -> List[Message]:
        if self.jinja:
            template = Template(self.function.__doc__, keep_trailing_newline=True)
            content = template.render(**inputs)
        else:
            content = self.function.__doc__.format(**inputs)

        import re

        pattern = r"-.*?(user|system|assistant):(.*?)(?=- \w+:|\Z)"
        matches = re.findall(pattern, content, re.DOTALL | re.MULTILINE)
        result = [(m[0], m[1].strip()) for m in matches]

        messages = []

        for role, content in result:
            if role not in ["user", "system", "assistant"]:
                raise BadRoleError(f"Role {role} is not valid")
            messages.append(Message(role=role, content=content.rstrip().lstrip()))

        return messages

    def call_llm(self, llm_data: LLMDataAndResult) -> LLMDataAndResult:
        if self.parser.llm_schema():  # pydantic schema
            return_scheme_llm_str = self.parser.llm_schema()
            ret_str = self.llm.call(llm_data.messages, scheme=return_scheme_llm_str)
        else:  # simple typings
            return_scheme_llm_str = self.parser.llm_return_type()
            ret_str = self.llm.call(
                llm_data.messages, return_type=return_scheme_llm_str
            )

        llm_data.raw_result = ret_str
        res = self.llm.clean_result(ret_str)
        llm_data.clean_result = res

        self.parser.cast_result(llm_data)
        logger.debug(f"Response from llm: \n{ret_str}")
        return llm_data


class Prompter:
    def __init__(self, llm: str, model_name: str, jinja=False, model_settings: Optional[Dict] = None):
        self.model_name = model_name
        self.llm = llm
        self.jinja = jinja
        self.model_settings = model_settings

    def __call__(self, function):
        return _Pr(
            function=function,
            jinja=self.jinja,
            llm=self.llm,
            model_name=self.model_name,
            model_settings=self.model_settings,
        )


import json
import random
from typing import List, Union

from pydantic_prompter.common import Message, logger
from pydantic_prompter.llm_providers.bedrock_base import BedRock


class BedRockCohere(BedRock):
    @property
    def _template_path(self) -> str:
        path = self.settings.template_paths.cohere.replace(
            "{prompt_paths}", self.parser.prompts_path
        )
        return path

    @property
    def _stop_sequence(self) -> str:
        return "User:"

    def format_messages(self, msgs: List[Message]) -> str:
        role_converter = {"user": "User", "system": "System", "assistant": "Chatbot"}
        output = []
        for msg in msgs:
            output.append(f"{role_converter[msg.role]}: {msg.content}")
        return "\n".join(output)

    def call(
        self,
        messages: List[Message],
        scheme: Union[dict, None] = None,
        return_type: Union[str, None] = None,
    ) -> str:
        content = self._build_prompt(messages, scheme or return_type)

        body = json.dumps(
            {
                "prompt": content,
                "stop_sequences": [self._stop_sequence],
                "temperature": random.uniform(0, 1),
            }
        )
        response = self._boto_invoke(body)

        response_body = json.loads(response.get("body").read().decode())
        logger.info(response_body)

        return response_body["generations"][0]["text"]



import logging

import pytest

from pydantic_prompter import Prompter
from pydantic_prompter.exceptions import (
    CohereAuthenticationError,
    OpenAiAuthenticationError,
    BedRockAuthenticationError,
)
from tests.data_for_tests import *

logger = logging.getLogger()

param_tests = pytest.mark.parametrize(
    "llm,model",
    [
        ("bedrock", "meta.llama2-70b-chat-v1"),
        # ("bedrock", "meta.llama2-13b-chat-v1"),
        ("bedrock", "cohere.command-text-v14"),
        ("openai", "gpt-3.5-turbo"),
        ("bedrock", "anthropic.claude-instant-v1"),
        ("bedrock", "anthropic.claude-v2"),
        ("bedrock", "anthropic.claude-3-sonnet-20240229-v1:0"),
        ("cohere", "command"),
        ("cohere", "command-light"),
    ],
)


@pytest.mark.live
@param_tests
def test_pydantic_result(llm, model):
    @Prompter(jinja=True, llm=llm, model_name=model)
    def bbb(name) -> PersonalInfo:
        """
        - user:
            hi, my name is {{ name }} and my children are called, aa, bb, cc
            what is my name and my children name
        """

    try:
        res: PersonalInfo = bbb(name="Ofer")
        assert isinstance(res, PersonalInfo)
        assert res.name == "Ofer"
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ):
        print(bbb.build_string())
        logger.exception("")
        pytest.skip("unsupported configuration")


@pytest.mark.live
@param_tests
def test_non_yaml_result(llm, model):
    @Prompter(llm=llm, model_name=model, jinja=True)
    def search_query(history) -> QueryGPTResponse:
        """
        {{ history }}

        - user:
            Generate a Google-like search query text encompassing all previous chat questions and answers
        """

    history = [
        "- user: Hi",
        "- assistant: what genre do you want to watch?",
        "- user: Comedy",
        "- assistant: do you want a movie or series?",
        "- user: Movie",
        "- assistant: OK",
    ]

    try:
        res = search_query(history="\n".join(history))
        assert isinstance(res, QueryGPTResponse)
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ) as e:
        logger.warning(e)
        pytest.skip("unsupported configuration")


@pytest.mark.live
@param_tests
def test_complex_question_result(llm, model):
    @Prompter(llm=llm, jinja=True, model_name=model)
    def rank_recommendation(json_entries, query) -> RecommendationResults:
        """
        - user:
            Which of the following JSON entries fit best to the query. order by best fit descending
            Base your answer ONLY on the given YML entries, if you are not sure, or there are no entries

            The JSON entries:
            {{ json_entries }}

            Query - {{ query }}

        """

    try:
        res = rank_recommendation(json_entries=entries, query=query)
        assert isinstance(res, RecommendationResults)
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ) as e:
        logger.warning(e)
        pytest.skip("unsupported configuration")


@pytest.mark.live
@param_tests
def test_str_result(llm, model):
    @Prompter(jinja=True, llm=llm, model_name=model)
    def bbb(name) -> str:
        """
        - user:
            hi, my name is {{ name }} and my children are called, aa, bb, cc
            what is my name and my children name
        """

    try:
        res = bbb(name="Ofer")
        assert isinstance(res, str)
        assert "Ofer" in res
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ):
        print(bbb.build_string())
        logger.exception("")
        pytest.skip("unsupported configuration")


@pytest.mark.live
@param_tests
def test_int_result(llm, model):
    @Prompter(jinja=True, llm=llm, model_name=model)
    def bbb(name) -> int:
        """
        - user:
            hi, my name is {{ name }} and my children are called, aa, bb, cc
            how many children do I have?
        """

    try:
        res = bbb(name="Ofer")
        print(res)
        assert isinstance(res, int)
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ):
        print(bbb.build_string())
        logger.exception("")
        pytest.skip("unsupported configuration")


@pytest.mark.live
@param_tests
def test_bool_result(llm, model):
    @Prompter(jinja=True, llm=llm, model_name=model)
    def bbb(name) -> bool:
        """
        - user:
            hi, my name is {{ name }} and my children are called, aa, bb, cc
            do I have children?
        """

    try:
        res = bbb(name="Ofer")
        print(res)
        assert isinstance(res, bool)
    except (
        OpenAiAuthenticationError,
        BedRockAuthenticationError,
        CohereAuthenticationError,
    ):
        print(bbb.build_string())
        logger.exception("")
        pytest.skip("unsupported configuration")



class NonRetryable(Exception):
    pass


class Retryable(Exception):
    pass


class OpenAiAuthenticationError(NonRetryable):
    pass


class BedRockAuthenticationError(NonRetryable):
    pass


class FailedToParsePromptError(NonRetryable):
    pass


class BadRoleError(NonRetryable):
    pass


class FailedToCastLLMResult(Retryable):
    pass


class CohereAuthenticationError(NonRetryable):
    pass


class ArgumentError(NonRetryable):
    pass



import logging
import pytest
from pydantic_prompter.exceptions import ArgumentError
from tests.data_for_tests import *
from pydantic_prompter import Prompter
from pydantic_prompter.prompter import Message


logging.getLogger("pydantic_prompter").setLevel(logging.DEBUG)
logging.basicConfig(
    level=logging.DEBUG,
)

logger = logging.getLogger("pydantic_prompter_test")


def test_basic():
    from pydantic_prompter import Prompter
    from pydantic import BaseModel

    class Hi(BaseModel):
        response: str

    @Prompter(llm="openai", jinja=False, model_name="gpt-3.5-turbo")
    def funct(hello) -> Hi:
        """
        - user: say {hello}
        """

    funct.build_string(hello="hi")


def test_basic_2():
    from pydantic_prompter import Prompter
    from pydantic import BaseModel

    logging.getLogger("pydantic_prompter").setLevel(logging.DEBUG)
    logging.basicConfig(
        level=logging.DEBUG,
    )

    class Hi(BaseModel):
        response: str

    @Prompter(llm="openai", jinja=False, model_name="gpt-3.5-turbo")
    def funct(hello) -> Hi:
        """
        - user: say {hello}
        """

    assert (
        funct.build_string(hello="hi")
        == '[\n    {\n        "content": "say hi",\n        "role": "user"\n    }\n]'
    )


def test_pydantic():
    @Prompter(jinja=True, llm="openai", model_name="gpt-3.5-turbo")
    def bbb(name) -> PersonalInfo:
        """
        - system: you are a writer
        - user: hi, my name is {{ name }} and my children are called, aa, bb, cc
        - user:
            what is my name and my children name
        """

    expected = [
        Message(role="system", content="you are a writer"),
        Message(
            role="user",
            content="hi, my name is Ofer and my children are called, aa, bb, cc",
        ),
        Message(role="user", content="what is my name and my children name"),
    ]
    res = bbb._parse_function_to_messages(name="Ofer")
    assert res == expected


def test_generic():
    @Prompter(llm="openai", model_name="dc")
    def aaa(name) -> MyChildren:
        """
        - user: hi, my name is {name} and my children are called, aa, bb, cc
        - user:
            how many children do I have and what's their names?
        """

    expected = [
        Message(
            role="user",
            content="hi, my name is Ofer and my children are called, aa, bb, cc",
        ),
        Message(
            role="user", content="how many children do I have and what's their names?"
        ),
    ]
    res = aaa._parse_function_to_messages(name="Ofer")
    assert res == expected


def test_non_yaml():
    @Prompter(llm="openai", model_name="dontcare", jinja=True)
    def search_query(history) -> QueryGPTResponse:
        """
        {{ history }}

        - user:
            Generate a Google-like search query text encompassing all previous chat questions and answers
        """

    history = [
        "- assistant: what genre do you want to watch?",
        "- user: Comedy",
        "- assistant: do you want a movie or series?",
        "- user: Movie",
    ]
    res = search_query._parse_function_to_messages(history="\n".join(history))
    expected = [
        Message(role="assistant", content="what genre do you want to watch?"),
        Message(role="user", content="Comedy"),
        Message(role="assistant", content="do you want a movie or series?"),
        Message(role="user", content="Movie"),
        Message(
            role="user",
            content="Generate a Google-like search query "
            "text encompassing all previous chat questions and answers",
        ),
    ]
    assert res == expected


def test_complex_question():
    @Prompter(llm="openai", jinja=True, model_name="dc")
    def rank_recommendation(json_entries, query) -> RecommendationResults:
        """
        - user:
            Which of the following JSON entries fit best to the query. order by best fit descending
            Base your answer ONLY on the given YML entries, if you are not sure, or there are no entries

        - user:
            The JSON entries:
            {{ json_entries }}

        - user: Query - {{ query }}

        """

    expected = [
        Message(
            role="user",
            content="Which of the following JSON entries fit best to the query. order by best fit descending\n            Base your answer ONLY on the given YML entries, if you are not sure, or there are no entries",
        ),
        Message(
            role="user",
            content='The JSON entries:\n            [{"text": "Description: 65 million years ago, the only 2 survivors of a spaceship from Somaris\\n  that crash-landed on Earth must fend off dinosaurs and reach the escape vessel in\\n  time before an imminent asteroid strike threatens to destroy the planet.\\nactors: Heath Ledger; Maggie Gyllenhaal; Aaron Eckhart; Christian Bale; Michael Caine;\\n  Gary Oldman\\ngenre: Action; drama; Thriller; Crime\\nid: 2054798\\nname: \'65\'\\nyear: 2008", "id": "2054798", "name": "65", "score": 0.4606173621133591}, {"text": "Description: \'\\"Selma,\\" as in Alabama, the place where segregation in the South was\\n  at its worst, leading to a march that ended in violence, forcing a famous statement\\n  by President Lyndon B. Johnson that ultimately led to the signing of the Voting\\n  Rights Act.\'\\nactors: Alessandro Nivola; Tom Wilkinson; Giovanni Ribisi; Carmen Ejogo; David Oyelowo;\\n  Cuba Gooding Jr.\\ngenre: A-Z; Biography;  Drama; Movies; History\\nid: 2312551\\nname: Selma\\nyear: 2014", "id": "2312551", "name": "Selma", "score": 0.4856962682297593}, {"text": "Description: 25-year-old Dwayne McLaren, a former athlete turned auto mechanic, dreams\\n  of getting out of tiny Cut Bank, Montana the coldest town in America. But his effort\\n  to do so sets in motion a deadly series of events that change his life and the life\\n  of the town forever...\\nactors: Michael Stuhlbarg; Billy Bob Thornton; John Malkovich; Teresa Palmer; Bruce\\n  Dern; Liam Hemsworth\\ngenre: SHO Movies; A-Z; Thriller\\nid: 2297150\\nname: Cut Bank\\nyear: 2015", "id": "2297150", "name": "Cut Bank", "score": 0.4997423205948504}, {"text": "Description: 2003. As a parasitic fungal outbreak begins to ravage the country and\\n  the world, Joel Miller attempts to escape the escalating chaos with his daughter\\n  and brother. Twenty years later, a now hardened Joel and his partner Tess fight\\n  to survive under a totalitarian regime, while the insurgent Fireflies harbor a teenage\\n  girl with a unique gift.\\nactors: Christopher Heyerdahl; Gabriel Luna; John Hannah; Marcia Bennett; Nico Parker;\\n  Brad Leland; Brendan Fletcher; Bella Ramsey; Pedro Pascal; Anna Torv; Merle Dandridge;\\n  Josh Brener; Jerry Wasserman\\ngenre: fantasy-sci-fi; drama; adventure; original; horror\\nid: 2044099\\nname: The Last of Us\\nyear: 2023", "id": "2044099", "name": "The Last of Us", "score": 0.5337261602153078}, {"text": "Description: \\"15 year-old Lyz, a high school student in the French Alps, has been\\\\\\n  \\\\ accepted to a highly selective ski club whose aim is to train future professional\\\\\\n  \\\\ athletes. Taking a chance on his new recruit, Fred, ex-champion turned coach,\\\\\\n  \\\\ decides to make Lyz his shining star regardless of her lack of experience. Under\\\\\\n  \\\\ his influence, Lyz will have to endure more than the physical and emotional pressure\\\\\\n  \\\\ of the training. Will Lyz\\\\u2019s determination help her escape his grip?\\"\\nactors: \\"Ma\\\\xEFra Schmitt; Marie Denarnaud; Axel Auriant; Muriel Combeau; J\\\\xE9r\\\\xE9\\\\\\n  mie Renier; No\\\\xE9e Abita\\"\\ngenre: SHO Movies; drama; A-Z\\nid: 2297161\\nname: Slalom\\nyear: 2021", "id": "2297161", "name": "Slalom", "score": 0.537035608231449}, {"text": "Description: \'\\"The Hours\\" is the story of three women searching for more potent, meaningful\\n  lives. Each is alive at a different time and place, all are linked by their yearnings\\n  and their fears. Their stories intertwine, and finally come together in a surprising,\\n  transcendent moment of shared recognition.\'\\nactors: Nicole Kidman; Julianne Moore; Meryl Streep; Ed Harris; Toni Collette; Claire\\n  Danes\\ngenre: A-Z; drama; SHO Movies\\nid: 2296757\\nname: The Hours\\nyear: 2002", "id": "2296757", "name": "The Hours", "score": 0.549504722004741}, {"text": "Description: \\"19-year-old Linn\\\\xE9a leaves her small town in Sweden and heads for\\\\\\n  \\\\ Los Angeles with the aim of becoming the world\'s next big porn star, but the road\\\\\\n  \\\\ to her goal turns out to be bumpier than she imagined.\\"\\nactors: Evelyn Claire; Alice Grey; Michael Omelko; Sofia Kappel; Jason Toler; Revika\\n  Reustle\\ngenre: drama; A-Z; SHO Movies\\nid: 2297023\\nname: Pleasure\\nyear: 2021", "id": "2297023", "name": "Pleasure", "score": 0.5547804101867869}, {"text": "Description: 70-year-old widower Ben Whittaker has discovered that retirement isn\'t\\n  all it\'s cracked up to be. Seizing an opportunity to get back in the game, he becomes\\n  a senior intern at an online fashion site, founded and run by Jules Ostin.\\nactors: JoJo Kushner; Anne Hathaway; Rene Russo; Anders Holm; Andrew Rannells; Robert\\n  De Niro\\ngenre: Movies; Comedy; A-Z\\nid: 2313618\\nname: The Intern (Theatrical)\\nyear: 2015", "id": "2313618", "name": "The Intern (Theatrical)", "score": 0.5573590045204179}, {"text": "Description: 27 years after overcoming the malevolent supernatural entity Pennywise,\\n  the former members of the Losers\' Club, who have grown up and moved away from Derry,\\n  are brought back together by a devastating phone call.\\nactors: James Ransone; Isaiah Mustafa; Jay Ryan; James McAvoy; Bill Hader; Jessica\\n  Chastain\\ngenre: drama; horror; Adaptation; A-Z; Fantasy; Movies\\nid: 2313729\\nname: \'It: Chapter Two\'\\nyear: 2019", "id": "2313729", "name": "It: Chapter Two", "score": 0.5695301178785966}, {"text": "Description: 20 years since their first adventure, Lloyd and Harry go on a road trip\\n  to find Harry\'s newly discovered daughter, who was given up for adoption.\\nactors: Steve Tom; Rachel Melvin; Laurie Holden; Jim Carrey; Jeff Daniels; Kathleen\\n  Turner\\ngenre: Movies; Comedy; A-Z\\nid: 2312803\\nname: Dumb and Dumber To\\nyear: 2014", "id": "2312803", "name": "Dumb and Dumber To", "score": 0.5716085222118849}]',
        ),
        Message(role="user", content="Query - Martial Arts, Action"),
    ]

    res = rank_recommendation._parse_function_to_messages(
        json_entries=entries, query=query
    )
    assert res == expected


def test_args_error():
    @Prompter(llm="cohere", model_name="command")
    def bbb(name) -> PersonalInfo:
        """
        - system: you are a writer
        - user:
            hi, my name is {{ name }} and my children are called, aa, bb, cc
            what is my name and my children name
        """

    with pytest.raises(ArgumentError) as e:
        bbb("Ofer")
        logger.info(str(e.value))
